/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_OGRCOORDINATETRANSFORMATION_HANDLE_HPP
#define GDAL_OGRCOORDINATETRANSFORMATION_HANDLE_HPP

#include <string>
#include <ogr_spatialref.h>

namespace gdal
{

  class OGRCoordinateTransformation_handle
  {
  public:
    using map_type = std::map<std::string, double>;
    using vec_type = std::vector<double>;

    OGRCoordinateTransformation_handle () = default;
    OGRCoordinateTransformation_handle (geographic_srs source, geographic_srs target)
      : p_(initialize(source, target))
    {}

    auto transform (double x, double y)
    {
      p_->Transform(1, &x, &y, NULL);
      return std::make_pair(x, y);
    }

    // Vector type coords arg must have 2 elements coresponding to
    // coords[0] = x coordinate, and coords[1] = y coordinate
    std::vector<double> operator () (std::vector<double> const& coords) const
      {
        assert(coords.size() == 2);

        double x = coords[0];
        double y = coords[1];
        p_->Transform(1, &x, &y, NULL);

        return std::vector<double> { x, y };
      }

    std::map<std::string, double> operator () (map_type coords) const
      {
        std::map<std::string, double> transformed_coords;

        double x = coords["x"];
        double y = coords["y"];
        p_->Transform(1, &x, &y, NULL);

        transformed_coords["x"] = x;
        transformed_coords["y"] = y;

        return transformed_coords;
      }

    auto operator () (std::pair<double, double> coords)
      {
        double x = coords.first;
        double y = coords.second;
        p_->Transform(1, &x, &y, NULL);
        return std::make_pair(x, y);
      }

    auto operator () (double x, double y)
      {
        p_->Transform(1, &x, &y, NULL);
        return std::make_pair(x, y);
      }

  private:
    using ptr_type = std::shared_ptr<OGRCoordinateTransformation>;

    struct deleter
    {
      void operator () (OGRCoordinateTransformation* p) const
      { OGRCoordinateTransformation::DestroyCT(p); }
    };

    ptr_type initialize (geographic_srs source, geographic_srs target)
    {
      auto p_raw = OGRCreateCoordinateTransformation(source.get(), target.get());
      ptr_type p(p_raw, deleter());
      return p;
    }

    ptr_type p_;
  };

} // namespace gdal

#endif
