/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE dataframe_range_proxy

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

#include "dataframe/range_proxy.hpp"
#include <string>
#include <vector>
#include <boost/mpl/vector.hpp>
#include <boost/range/algorithm/equal.hpp>

using test_types
= boost::mpl::vector<
  std::vector<int>
  , std::vector<long>
  , std::vector<long long>
  , std::vector<float>
  , std::vector<double>
  , std::vector<long double>
  >;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_typedef,Range,test_types)
{
  using range_proxy_type = dataframe::range_proxy<Range>;
  using const_range_proxy_type = dataframe::range_proxy<Range const>;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_constructor,Range,test_types)
{
  using range_proxy_type = dataframe::range_proxy<Range>;
  using const_range_proxy_type = dataframe::range_proxy<Range const>;

  range_proxy_type range_proxy;
  BOOST_CHECK(range_proxy.empty());
  BOOST_CHECK_EQUAL(range_proxy.size(),0);
  const_range_proxy_type const_range_proxy;
  BOOST_CHECK(const_range_proxy.empty());
  BOOST_CHECK_EQUAL(const_range_proxy.size(),0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_range_proxy,Range,test_types)
{
  using range_proxy_type = dataframe::range_proxy<Range>;

  using boost::equal;
  
  Range range = { 42, 17 };
  range_proxy_type range_proxy(range);
  BOOST_CHECK(!range_proxy.empty());
  BOOST_CHECK_EQUAL(range_proxy.size(),2);
  BOOST_CHECK(equal(range_proxy,range));
  range_proxy = { 13 };
  BOOST_CHECK(!range_proxy.empty());
  BOOST_CHECK_EQUAL(range_proxy.size(),1);
  BOOST_CHECK(equal(range_proxy,Range({13})));
  range_proxy = { 17 };
  BOOST_CHECK(!range_proxy.empty());
  BOOST_CHECK_EQUAL(range_proxy.size(),1);
  BOOST_CHECK(equal(range_proxy,Range(1,17)));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_const_range_proxy,Range,test_types)
{
  using range_proxy_type = dataframe::range_proxy<Range>;
  using const_range_proxy_type = dataframe::range_proxy<Range const>;

  using boost::equal;
  
  Range range = { 42, 17 };
  const_range_proxy_type range_proxy(range);
  BOOST_CHECK(!range_proxy.empty());
  BOOST_CHECK_EQUAL(range_proxy.size(),2);
  BOOST_CHECK(equal(range_proxy,Range({42,17})));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_range_proxy,Range,test_types)
{
  using range_proxy_type = dataframe::range_proxy<Range>;
  using value_type = typename Range::value_type;
  
  using dataframe::make_range_proxy;
  
  Range range = { 42, 17 };
  auto range_proxy = make_range_proxy<value_type>(range);
  static_assert(std::is_same<decltype(range_proxy),range_proxy_type>::value,"");
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_const_range_proxy,Range,test_types)
{
  using const_range_proxy_type = dataframe::range_proxy<Range const>;
  using value_type = typename Range::value_type;

  using dataframe::make_range_proxy;
  
  Range const range = { 42, 17 };
  auto const_range_proxy = make_range_proxy<value_type>(range);
  static_assert(std::is_same<decltype(const_range_proxy),const_range_proxy_type>::value,"");
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_range_iterator,Range,test_types)
{
  using value_type = typename Range::value_type;
  using dataframe::make_range_proxy;

  Range range = { 42, 17 };
  auto range_proxy = make_range_proxy<value_type>(range);
  BOOST_CHECK_EQUAL(*begin(range_proxy),42);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_range_iterator_2,Range,test_types)
{
  using dataframe::make_range_proxy;

  Range range = { 42, 17 };
  auto range_proxy = make_range_proxy<std::string>(range);
  BOOST_CHECK_EQUAL(*begin(range_proxy),"42");
}

