/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE gdal

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <iostream>
#include <boost/mpl/vector.hpp>
#include <iostream>
#include <ogr_core.h>
#include <gdal/rasterdataset.hpp>
#include <gdal/point.hpp>
#include <gdal/geographic_srs.hpp>
#include <gdal/grid_srs.hpp>
#include <gdal/affine_transform.hpp>


using std::cout;
using std::endl;
using namespace gdal;

#define rasterfile "../data/test_raster.tif"


// point<geographic_srs>
BOOST_AUTO_TEST_CASE(point_geographic_default_constructor)
{
  point<geographic_srs>();
}

BOOST_AUTO_TEST_CASE(geographic_point_srs_constructor)
{
  double x(-107.01599);
  double y(  32.31111);
  auto wgs84_srs = make_wgs84();

  point<geographic_srs> pt(x, y, wgs84_srs);
}

BOOST_AUTO_TEST_CASE(geographic_point_copy_constructor)
{
  double x(-107.01599);
  double y(  32.31111);
  auto wgs84_srs = make_wgs84();

  point<geographic_srs> pt(x, y, wgs84_srs);
  point<geographic_srs> pt2(pt);
}

BOOST_AUTO_TEST_CASE(geographic_point_move_constructor)
{
  double x(-107.01599);
  double y(  32.31111);
  auto wgs84_srs = make_wgs84();

  point<geographic_srs> pt(x, y, wgs84_srs);
  point<geographic_srs> pt2 = std::move(pt);
}

BOOST_AUTO_TEST_CASE(geographic_point_x)
{
double x = -1030133;
double y = 1086599;
auto raster_srs = rasterdataset<>(rasterfile).srs();
point<geographic_srs> pt(x, y, raster_srs);
BOOST_CHECK_EQUAL(pt.x(), x);
}

BOOST_AUTO_TEST_CASE(geographic_point_y)
{
double x = -1030133;
double y = 1086599;
auto raster_srs = rasterdataset<>(rasterfile).srs();
point<geographic_srs> pt(x, y, raster_srs);
BOOST_CHECK_EQUAL(pt.y(), y);
}

BOOST_AUTO_TEST_CASE(geographic_point_equality_comparable)
{
  auto srs1 = make_wgs84();
  auto srs2 = rasterdataset<>(rasterfile).srs();

  double x1(-107.0); double y1(  32.0);
  double x2(-107.1); double y2(  32.1);

  point<geographic_srs> pt1(x1, y1, srs1);
  point<geographic_srs> pt1_copy(pt1);
  point<geographic_srs> pt2(x2, y1, srs1);
  point<geographic_srs> pt3(x1, y2, srs1);
  point<geographic_srs> pt4(x1, y1, srs2);
  point<geographic_srs> pt5(x2, y2, srs2);

  BOOST_CHECK_EQUAL(pt1, pt1_copy);
  BOOST_CHECK(pt1 != pt2);
  BOOST_CHECK(pt1 != pt3);
  BOOST_CHECK(pt1 != pt4);
  BOOST_CHECK(pt1 != pt5);
}

// point<grid_srs>
BOOST_AUTO_TEST_CASE(grid_point_default_constructor)
{
  point<grid_srs> pt;
}

BOOST_AUTO_TEST_CASE(grid_point_coord_constructor)
{
  int x(3); int y(5);
  point<grid_srs> pt(x,y);
}

BOOST_AUTO_TEST_CASE(grid_point_coord_and_srs_constructor)
{
 int x(3);int y(5);
  grid_srs grid(4,7);
  point<grid_srs> pt(x,y, grid);
}

BOOST_AUTO_TEST_CASE(grid_point_copy_constructor)
{
 int x(3);int y(5);
  grid_srs grid(4,7);
  point<grid_srs> pt(x,y, grid);

  point<grid_srs> pt_copy(pt);
}

BOOST_AUTO_TEST_CASE(grid_point_move_constructor)
{
 int x(3);int y(5);
  grid_srs grid(4,7);
  point<grid_srs> pt(x,y, grid);

  auto pt_moved = std::move(pt);
}

BOOST_AUTO_TEST_CASE(grid_point_x)
{
 int x(3);int y(5);
  grid_srs grid(4,7);
  point<grid_srs> pt(x,y, grid);

  BOOST_CHECK_EQUAL(pt.x(), x);
}

BOOST_AUTO_TEST_CASE(grid_point_y)
{
 int x(3);int y(5);
  grid_srs grid(4,7);
  point<grid_srs> pt(x,y, grid);

  BOOST_CHECK_EQUAL(pt.y(), y);
}

BOOST_AUTO_TEST_CASE(grid_point_equality_comparable)
{
 int x1(3);int y1(5);
 int x2(4);int y2(1);

  grid_srs grid1(4,7);
  grid_srs grid2(5,7);

  point<grid_srs> pt1(x1, y1, grid1);
  point<grid_srs> pt1_copy(pt1);
  BOOST_CHECK_EQUAL(pt1, pt1_copy);

  point<grid_srs> pt2(x2, y1, grid1);
  point<grid_srs> pt3(x1, y2, grid1);
  point<grid_srs> pt4(x1, y1, grid2);

  BOOST_CHECK(pt1 != pt2);
  BOOST_CHECK(pt1 != pt3);
  BOOST_CHECK(pt1 != pt4);
}
