/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_distribution_bernoulli_bernoulli_distribution

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <graph_model/distribution/bernoulli/bernoulli_distribution.hpp>
#include <graph_model/distribution/parameter/probability.hpp>
#include <graph_model/distribution/parameter/variate.hpp>
#include <graph_model/vertex/lit.hpp>

#if XXX_TOO_MANY_ARGUMENTS
BOOST_AUTO_TEST_CASE(test_too_many_arguments)
{
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  auto distribution = bernoulli_distribution(1,2,3);
}
#endif

BOOST_AUTO_TEST_CASE(test_default_constructor)
{
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  using graph_model::distribution::parameter::probability;
  using graph_model::distribution::parameter::variate;
  {
    auto distribution = bernoulli_distribution();
    BOOST_CHECK_EQUAL(distribution.p(),0.5);
    BOOST_CHECK_EQUAL(distribution.variate(),true);
  }
  {
    auto distribution = bernoulli_distribution(0.25);
    BOOST_CHECK_EQUAL(distribution.p(),0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),true);
  }
  {
    auto distribution = bernoulli_distribution(0.25,false);
    BOOST_CHECK_EQUAL(distribution.p(),0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
  {
    auto distribution = bernoulli_distribution(probability=0.25);
    BOOST_CHECK_EQUAL(distribution.p(),0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),true);
  }
  {
    auto distribution = bernoulli_distribution(probability=0.25,false);
    BOOST_CHECK_EQUAL(distribution.p(),0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
  {
    auto distribution = bernoulli_distribution(false,probability=0.25);
    BOOST_CHECK_EQUAL(distribution.p(),0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
  {
    auto distribution = bernoulli_distribution(probability=0.25,variate=false);
    BOOST_CHECK_EQUAL(distribution.p(),0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
}

BOOST_AUTO_TEST_CASE(test_constructor)
{
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  {
    auto distribution = bernoulli_distribution();
    BOOST_CHECK_EQUAL(distribution.p(),0.5);
    BOOST_CHECK_EQUAL(distribution.variate(),true);
  }
  {
    auto distribution = bernoulli_distribution(0);
    BOOST_CHECK_EQUAL(distribution.p(),0);
    BOOST_CHECK_EQUAL(distribution.variate(),true);
  }
  {
    auto distribution = bernoulli_distribution(0.25,false);
    BOOST_CHECK_EQUAL(distribution.p(),0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
}

BOOST_AUTO_TEST_CASE(test_constructor_lit)
{
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  using graph_model::vertex::lit;

  auto p = lit(0.25);
  auto distribution = bernoulli_distribution(p,false);
  BOOST_CHECK_EQUAL(distribution.p(),0.25);
  BOOST_CHECK_EQUAL(distribution.variate(),false);
}

BOOST_AUTO_TEST_CASE(test_constructor_expr)
{
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  using graph_model::vertex::lit;

  auto variate = false;
  auto p = lit(0) + 0.25;
  auto distribution = bernoulli_distribution(p,variate);
  BOOST_CHECK_EQUAL(distribution.p(),0 + 0.25);
  BOOST_CHECK_EQUAL(distribution.variate(),false);
}

BOOST_AUTO_TEST_CASE(test_constructor_tagged_parameters)
{
  using std::sqrt;

  using graph_model::distribution::bernoulli::bernoulli_distribution;
  using graph_model::distribution::parameter::probability;
  using graph_model::distribution::parameter::variate;
  using graph_model::vertex::lit;

  auto v = false;
  auto p = lit(0) + 0.25;

  {
    auto distribution = bernoulli_distribution(variate=v,probability=p);
    BOOST_CHECK_EQUAL(distribution.p(),0 + 0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
  {
    auto distribution = bernoulli_distribution(v,probability=p);
    BOOST_CHECK_EQUAL(distribution.p(),0 + 0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
  {
    auto distribution = bernoulli_distribution(variate=v,p);
    BOOST_CHECK_EQUAL(distribution.p(),0 + 0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
}

BOOST_AUTO_TEST_CASE(test_constructor_tagged_parameter_pack)
{
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  using graph_model::distribution::parameter::probability;
  using graph_model::distribution::parameter::variate;
  using graph_model::vertex::lit;

  auto v = false;
  auto p = lit(0) + 0.25;

  {
    auto distribution = bernoulli_distribution(variate=v,probability=p);
    BOOST_CHECK_EQUAL(distribution.p(),0 + 0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
  {
    auto distribution = bernoulli_distribution(variate=v,p);
    BOOST_CHECK_EQUAL(distribution.p(),0 + 0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
  {
    auto distribution = bernoulli_distribution(probability=p,v);
    BOOST_CHECK_EQUAL(distribution.p(),0 + 0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
  {
    auto distribution = bernoulli_distribution(p,v);
    BOOST_CHECK_EQUAL(distribution.p(),0 + 0.25);
    BOOST_CHECK_EQUAL(distribution.variate(),false);
  }
}
