/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_examples_normal_sample_mean_truncated

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#if EXECUTE_ASYNCHRONOUSLY
#include <thread>
#include <boost/asynchronous/queue/lockfree_queue.hpp>
#include <boost/asynchronous/scheduler/multiqueue_threadpool_scheduler.hpp>
#include <boost/asynchronous/scheduler/single_thread_scheduler.hpp>
#include <boost/asynchronous/scheduler_shared_proxy.hpp>
#include <asynchronous/signals/type/default_signal_executor_is_parallel.hpp>
#include <asynchronous/signals/type/job.hpp>
#else  // EXECUTE_ASYNCHRONOUSLY
#include <asynchronous/signals/type/default_signal_executor_is_sequential.hpp>
#endif // EXECUTE_ASYNCHRONOUSLY

#include <iostream>
#include <map>

#include <boost/filesystem/fstream.hpp>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>

#include <asynchronous/signals/signal_diagnostics.hpp>

#include <dataframe/dataframe.hpp>
#include <dataframe/element.hpp>
#include <dataframe/read_dataframe.hpp>

#include <graph_model/MCMC/mcmc_signal_policy.hpp>
#include <graph_model/MCMC/metropolis_sampler.hpp>
#include <graph_model/accumulators/accumulator_set.hpp>
#include <graph_model/distribution/normal/normal_distribution.hpp>
#include <graph_model/graph/graphviz.hpp>
#include <graph_model/graph_model.hpp>
#include <graph_model/signals/tags.hpp>
#include <graph_model/vertex/cdf.hpp>
#include <graph_model/vertex/expression.hpp>
#include <graph_model/vertex/factor.hpp>
#include <graph_model/vertex/lit.hpp>
#include <graph_model/vertex/markov_blanket.hpp>
#include <graph_model/vertex/pow.hpp>
#include <graph_model/vertex/random_variable.hpp>
#include <graph_model/vertex/sample.hpp>

#include <identifiers/x.hpp>
#include <mcmc/mcmc_tally.hpp>

#include <logging/initialize.hpp>

#include <boost/mpl/pair.hpp>
#include <boost/mpl/vector.hpp>

// Use a lognumeric type to represent probabilities
//
#if USE_LOGNUMERIC_PROBABILITY_TYPE
#include <sample_space/numeric/lognumeric.hpp>
#endif	// USE_LOGNUMERIC_PROBABILITY_TYPE

using test_types = boost::mpl::vector<
  boost::mpl::pair<double,long double>
  >;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_graph_model,Pair,test_types)
{
  using variate_type = typename Pair::first;
#if USE_LOGNUMERIC_PROBABILITY_TYPE
  using probability_type = sample_space::numeric::lognumeric<typename Pair::second>;
#else  // USE_LOGNUMERIC_PROBABILITY_TYPE
  using probability_type = typename Pair::second;
#endif	// USE_LOGNUMERIC_PROBABILITY_TYPE

  using ifstream_type = boost::filesystem::ifstream;
  using ofstream_type = boost::filesystem::ofstream;

  using factor_type = graph_model::vertex::factor<probability_type>;
  using factor_map_type = std::map<int,factor_type>;

  using expression_type = graph_model::vertex::expression<variate_type>;
  using data_map_type = std::map<int,expression_type>;

  using iteration = graph_model::signals::tag::iteration;
  using accept = graph_model::signals::tag::accept;
  using reject = graph_model::signals::tag::reject;

  using severity_level = boost::log::trivial::severity_level;
  using logger_type = boost::log::sources::severity_logger<severity_level>;

  using graph_model::vertex::lit;

  using dataframe::make_dataframe;
  using dataframe::make_element;

  using graph_model::MCMC::make_mcmc_signal_policy;
  using graph_model::MCMC::metropolis_sampler;
  using graph_model::accumulators::make_accumulator_set;
  using graph_model::distribution::normal::normal_distribution;
  using graph_model::make_incidence_graph;
  using graph_model::make_vertex_and_edge_list_graph;
  using graph_model::vertex::make_random_variable;

  using graph_model::distribution::parameter::mean;
  using graph_model::distribution::parameter::standard_deviation;
  using graph_model::distribution::parameter::variate;

#if EXECUTE_ASYNCHRONOUSLY
  using job_type = asynchronous::signals::type::default_job_t<>;
  using queue_type = boost::asynchronous::lockfree_queue<job_type>;
  using find_position_type = boost::asynchronous::default_find_position<>;
  using cpu_load_type = boost::asynchronous::no_cpu_load_saving;
  using scheduler_type = boost::asynchronous::single_thread_scheduler<queue_type,cpu_load_type>;
  using threadpool_type = boost::asynchronous::multiqueue_threadpool_scheduler<queue_type,find_position_type,cpu_load_type>;

  using boost::asynchronous::make_shared_scheduler_proxy;
#endif // EXECUTE_ASYNCHRONOUSLY

  using identifiers::x;
  using mcmc::make_mcmc_tally;

  namespace accumulators = boost::accumulators;

  logging::initialize();
  logger_type debug_log;

#if EXECUTE_ASYNCHRONOUSLY
  auto scheduler = make_shared_scheduler_proxy<scheduler_type>();
  auto threadpool = make_shared_scheduler_proxy<threadpool_type>(std::thread::hardware_concurrency()/2);
#endif // EXECUTE_ASYNCHRONOUSLY

  // construct the random variable to estimate
  auto mu = make_random_variable<variate_type,probability_type>
    (
#if EXECUTE_ASYNCHRONOUSLY
     scheduler,threadpool
#endif // EXECUTE_ASYNCHRONOUSLY
     );
  mu.reset_identifier("mu");

  // construct a dataframe of observations
  ifstream_type is("samples.dat");
  auto dataframe = make_dataframe();
  is >> dataframe;

  // construct random variables for the observations and corresponding factors
  data_map_type data;
  factor_map_type factor;
  for (auto row : dataframe)
    {
      data[row].reset_identifier(x(row));
      data[row].reset_strategy
	(make_element<variate_type>(dataframe,row,"samples"));
      factor[row]
	.reset_identifier(row)
	.reset_strategy(normal_distribution(variate=data[row],mean=mu,standard_deviation=1))
	.markov_blanket(mu);
      data[row].add_vertex(factor[row]);
      // mu.add_factor(lit(probability_type(1))/(graph_model::vertex::cdf(normal_distribution(mu,1),2)-cdf(normal_distribution(mu,1),-2)));
      // mu.add_factor(pow(lit(probability_type(1))/(graph_model::vertex::cdf(normal_distribution(mu,1),2)-cdf(normal_distribution(mu,1),-2)),1));
    }
  mu.add_factor(pow(lit(probability_type(1))/(graph_model::vertex::cdf(normal_distribution(mu,1),2)-cdf(normal_distribution(mu,1),-2)),dataframe.size()));

  // construct the MCMC components
  auto accumulator = make_accumulator_set<variate_type>();
  auto mcmc_tally = make_mcmc_tally();
  auto mu_sampler
    = metropolis_sampler
    (markov_blanket(mu),
     make_mcmc_signal_policy
     (
#if EXECUTE_ASYNCHRONOUSLY
      scheduler,threadpool
#endif // EXECUTE_ASYNCHRONOUSLY
      )
     .template connect_signal<iteration>
     ([]{ static std::size_t i = 0; if (i == 0) std::cout << "iteration: "; if (i % 100 == 0) std::cout << " " << i << std::flush; ++i; })
     .template connect_signal<accept>(mcmc_tally.make_acceptor())
     .template connect_signal<reject>(mcmc_tally.make_rejector()))
    .add_rv(mu,lit(mu) + sample(normal_distribution(0,2)));
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "mu_sampler.initialize()";
  mu_sampler.initialize();
  mu_sampler.commit();
  accumulator(mu.variate());

  // construct an output stream for the MCMC samples and generate samples
  ofstream_type ofstream("mcmc.dat");
  ofstream << "i\tMCMC_sample\tmean" << std::endl;
  ofstream << 0 << "\t" << mu.variate() << "\t" << accumulators::mean(accumulator) << std::endl;
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC samples:";
  for (int i = 0; i < MCMC_SAMPLE_SIZE; ++i)
    {
      GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC sample: " << i;
      mu_sampler.update();
      mu_sampler.commit();
      accumulator(mu.variate());
      ofstream << i+1 << "\t" << mu.variate() << "\t" << accumulators::mean(accumulator) << std::endl;
    }
  std::cout << std::endl;

  // report summary statistics
  std::cout << "estimated mu:  " << accumulators::mean(accumulator) << std::endl;
  std::cout << "MCMC tally:    " << mcmc_tally << std::endl;

#if EXECUTE_ASYNCHRONOUSLY
  // report asynchronous diagnostics  
  using asynchronous::signals::signal_diagnostics;
  ofstream_type diagnostics_ostream("diagnostics.dat");
  diagnostics_ostream << signal_diagnostics::header() << std::endl;
  diagnostics_ostream << signal_diagnostics("scheduler",scheduler.get_diagnostics().totals());
  diagnostics_ostream << signal_diagnostics("threadpool",threadpool.get_diagnostics().totals());
#endif // EXECUTE_ASYNCHRONOUSLY
}
